// FTPLogin.cpp : Implementation of CFTPLogin
#include "stdafx.h"
#include "ActiveMail.h"
#include "FTPLogin.h"

/////////////////////////////////////////////////////////////////////////////
// CFTPLogin


STDMETHODIMP CFTPLogin::get_ServerName(BSTR * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  *pVal = T2BSTR(ErrorMessage);
  return S_OK;
}

STDMETHODIMP CFTPLogin::put_ServerName(BSTR newVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  LPSTR lpszNewVal = OLE2T(newVal);
  free(lpszServerName);
  lpszServerName = (char *) malloc(lstrlen(lpszNewVal) + 1);
  lstrcpy(lpszServerName, lpszNewVal);
  return S_OK;
}

STDMETHODIMP CFTPLogin::get_ErrorCode(long * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  *pVal = ErrorCode;
  return S_OK;
}

STDMETHODIMP CFTPLogin::get_ErrorMessage(BSTR * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  *pVal = T2BSTR(ErrorMessage);
  return S_OK;
}

STDMETHODIMP CFTPLogin::get_Authenticate(BSTR UserName, BSTR Password, BOOL * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  
  // Connect to the FTP server
  CSocket pSocket;
  int code;
  *pVal = false;
  pSocket.Create(0, SOCK_STREAM, 0);
  if (!pSocket.Connect(lpszServerName, 21)) {
	if (pSocket.GetLastError() == WSAENOTSOCK) {
	  pSocket.Close();
	  return EmitErrorMessage("Connection would block", 5);
	} else {
	  pSocket.Close();
 	  return EmitErrorMessage("Could not connect to server", 6);
	}
  }
  if ((code = GetResponseCode(&pSocket)) != 220) {
	pSocket.Close();
	return EmitErrorMessage("Server returned incorrect response on connection", code);
  }

  // Send the USER command
  LPTSTR lpszUserName = OLE2T(UserName);
  CString pData;
  pData = "USER ";
  pData += lpszUserName;
  pData += "\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send USER to the server", 7);
  }
  if ((code = GetResponseCode(&pSocket)) != 331) {
	pSocket.Close();
	return EmitErrorMessage("Server returned USER error", code);
  }

  // Send the PASS command
  LPTSTR lpszPassword = OLE2T(Password);
  pData = "PASS ";
  pData += lpszPassword;
  pData += "\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send PASS to the server", 8);
  }
  switch (code = GetResponseCode(&pSocket))
  {

  case 230 :
	{
	  // Login successful :-)
	  *pVal = true;
	  break;
	}

  case 530 :
	{
	  // Login failed :-(
      pSocket.Close();
	  return EmitErrorMessage("Access denied", code);
	}

  default :
	{
	  // Unknown login failure
	  pSocket.Close();
	  return EmitErrorMessage("Unknown login failure", code);
	}
  }
  
  // Send the QUIT command
  pData = "QUIT\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send QUIT to the server", 12);
  }
  if ((code = GetResponseCode(&pSocket)) != 221) {
	pSocket.Close();
	return EmitErrorMessage("Server returned QUIT error", code);
  }
  
  // Close the socket and make sure the return value is correct
  pSocket.Close();
  if (!*pVal)
	return EmitErrorMessage("Unknown error during processing!  This should never appear :-(", 13);
  return EmitErrorMessage("Login successful", 0);
}

BOOL CFTPLogin::EmitErrorMessage(LPCTSTR message, int code)
{
  // Store the error message and return S_OK
  free(ErrorMessage);
  ErrorMessage = (char *) malloc(lstrlen(message) + 1);
  lstrcpy(ErrorMessage, message);
  ErrorCode = code;
  return S_OK;
}

int CFTPLogin::GetResponseCode(CSocket * pSocket)
{
  char nextChar = ' ';
  char inputBuffer[256];
  int nextIndex = 0;
  
  while ((nextChar != '\n') && (nextIndex < 255)) {
    if (pSocket->Receive(&nextChar, 1, 0) == SOCKET_ERROR)
	  return -1;
	inputBuffer[nextIndex++] = nextChar;
  }
  inputBuffer[nextIndex] = '\0';
  
  // This is to support multi-line responses, common with
  // most FTP servers
  if(inputBuffer[3] == '-' )
	return GetResponseCode(pSocket);
  else
    return atoi(inputBuffer);
}
