// POP3Connection.cpp : Implementation of CPOP3Connection
#include "stdafx.h"
#include "ActiveMail.h"
#include "POP3Connection.h"

/////////////////////////////////////////////////////////////////////////////
// CPOP3Connection

// POP3: RFC 1725

// Timeout is always >= 10 minutes, if any

BOOL CPOP3Connection::EmitErrorMessage(LPCTSTR message, int code)
{
  // Store the error message and return S_OK
  free(ErrorMessage);
  ErrorMessage = (char *) malloc(lstrlen(message) + 1);
  lstrcpy(ErrorMessage, message);
  ErrorCode = code;
  return S_OK;
}

BOOL CPOP3Connection::ReadResponse(LPBOOL lpbOk)
{
  // Read the "+OK" or "-ERR" response
  char status;
  char nextChar = ' ';
  if (pSocket.Receive(&status, 1, 0) == SOCKET_ERROR)
    return false;
  
  // Read the rest of the line into a string
  int nextIndex = 0;
  inputBuffer = "";
  if (pSocket.Receive(&nextChar, 1, 0) == SOCKET_ERROR)
	return false;
  while (nextChar != '\n') {
	inputBuffer += nextChar;
	nextIndex++;
	if (pSocket.Receive(&nextChar, 1, 0) == SOCKET_ERROR)
	  return false;
  }
  if (inputBuffer.GetAt(nextIndex - 1) == '\r')
	inputBuffer.SetAt(nextIndex - 1, 0);

  // If the status was OK, then return true :-)
  if (status == '+') {
	*lpbOk = true;
	return true;
  }

  // If the status was ERR, then set the error message and return
  if (status == '-') {
	EmitErrorMessage(inputBuffer, 999);
	*lpbOk = false;
	return true;
  }

  // Uhh.. the server didn't return + or -!
  return false;
}

BOOL CPOP3Connection::ReadServerLine(LPBOOL lpbEnd)
{
  // Read the line into a string
  char nextChar = ' ';
  int nextIndex = 0;
  inputBuffer = "";
  if (pSocket.Receive(&nextChar, 1, 0) == SOCKET_ERROR)
	return false;
  while (nextChar != '\n') {
	inputBuffer += nextChar;
	nextIndex++;
	if (pSocket.Receive(&nextChar, 1, 0) == SOCKET_ERROR)
	  return false;
  }
  if (inputBuffer.GetAt(nextIndex - 1) == '\r')
	inputBuffer.SetAt(nextIndex - 1, 0);

  // If the whole line was just a period, then set lpbEnd
  if (inputBuffer == ".")
	*lpbEnd = true;
  else
	*lpbEnd = false;
  return true;
}

STDMETHODIMP CPOP3Connection::get_ErrorCode(long * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  *pVal = ErrorCode;
  return S_OK;
}

STDMETHODIMP CPOP3Connection::get_ErrorMessage(BSTR * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  *pVal = T2BSTR(ErrorMessage);
  return S_OK;
}


STDMETHODIMP CPOP3Connection::get_ServerName(BSTR * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  *pVal = T2BSTR(ErrorMessage);
  return S_OK;
}

STDMETHODIMP CPOP3Connection::put_ServerName(BSTR newVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  LPSTR lpszNewVal = OLE2T(newVal);
  free(lpszServerName);
  lpszServerName = (char *) malloc(lstrlen(lpszNewVal) + 1);
  lstrcpy(lpszServerName, lpszNewVal);
  return S_OK;
}

STDMETHODIMP CPOP3Connection::Connect(BSTR UserName, BSTR Password)
{
  // Connects to the server
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;

  // If we're connected, then close first
  if (connected) {
    this->Close();
  }

  // Convert strings
  LPTSTR lpszUserName = OLE2T(UserName);
  LPTSTR lpszPassword = OLE2T(Password);

  // Free up any existing allocations
  if (msgData) {
	DWORD i;
	for (i = 0; i < nMessages; i++) {
	  if (msgData[i] != NULL)
		delete msgData[i];
	}
	free(msgData);
	msgData = NULL;
  }
  if (lpdwMessageLengths) {
	free(lpdwMessageLengths);
	lpdwMessageLengths = NULL;
  }
  if (lpdwMessageIDs) {
	free(lpdwMessageIDs);
	lpdwMessageIDs = NULL;
  }

  // Connect to the POP3 server
  BOOL bOk;
  pSocket.Create(0, SOCK_STREAM, 0);
  if (!pSocket.Connect(lpszServerName, 110)) {
	if (pSocket.GetLastError() == WSAENOTSOCK) {
	  pSocket.Close();
	  return EmitErrorMessage("Connection would block", 5);
	} else
 	  return EmitErrorMessage("Could not connect to server", 6);
  }
  if (!ReadResponse(&bOk)) {
	pSocket.Close();
	return EmitErrorMessage("Remote host is not a POP3 server", 7);
  }
  if (!bOk)
  {
	// Server returned an error; perhaps there are too many connections at this time?
	pSocket.Close();
	return EmitErrorMessage((LPCTSTR) inputBuffer, 8);
  }

  // Send the username
  CString pData;
  pData = "USER ";
  pData += lpszUserName;
  pData += "\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send USER to the server", 9);
  }
  if (!ReadResponse(&bOk)) {
	pSocket.Close();
	return EmitErrorMessage("Server returned unexpected data on USER command", 10);
  }
  if (!bOk)
  {
	// Server returned an error; perhaps the user doesn't exist?
	pSocket.Close();
	return EmitErrorMessage((LPCTSTR) inputBuffer, 11);
  }

  // Send the password
  pData = "PASS ";
  pData += lpszPassword;
  pData += "\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send PASS to the server", 12);
  }
  if (!ReadResponse(&bOk)) {
	pSocket.Close();
	return EmitErrorMessage("Server returned unexpected data on PASS command", 13);
  }
  if (!bOk)
  {
	// Server returned an error; perhaps the user doesn't exist or the password was bad?
	pSocket.Close();
	return EmitErrorMessage((LPCTSTR) inputBuffer, 14);
  }

  // Send the STAT command to initialize the nMessages and dwMailboxLength variables
  pData = "STAT\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send STAT to the server", 15);
  }
  if (!ReadResponse(&bOk)) {
	pSocket.Close();
	return EmitErrorMessage("Server returned unexpected data on STAT command", 16);
  }

  // We're staying connected from here on...
  connected = true;

  // Continue to check STAT response
  if (!bOk)
  {
	// Was the mailbox locked?  In any case, set the variables to zero and return
	nMessages = 0;
	dwMailboxLength = 0;
	return S_OK;
  }

  // Read the variables
  if (sscanf((LPCTSTR) inputBuffer, "OK %lu %lu", &nMessages, &dwMailboxLength) != 2)
  {
	// Could not chew the line; emit an error
	nMessages = 0;
	dwMailboxLength = 0;
	return EmitErrorMessage("Could not decipher server response to STAT command", 17);
  }

  // Get the message lengths if messages exist
  if (nMessages) {
	lpdwMessageLengths = (LPDWORD) malloc(sizeof(DWORD) * nMessages);
	lpdwMessageIDs = (LPDWORD) malloc(sizeof(DWORD) * nMessages);
	msgData = (CString **) malloc(sizeof(CString **) * nMessages);

    // Send the LIST command to the server
    pData = "LIST\r\n";
    if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	  pSocket.Close();
	  return EmitErrorMessage("Could not send LIST to the server", 18);
	}
    if (!ReadResponse(&bOk)) {
	  pSocket.Close();
	  return EmitErrorMessage("Server returned unexpected data on LIST command", 19);
	}

	// Loop through the server data
	BOOL bEnd = false;
	DWORD i = 0;
	while (!bEnd) {
	  if (!ReadServerLine(&bEnd)) {
	    pSocket.Close();
	    return EmitErrorMessage("Server returned unexpected data on LIST command data", 20);
	  }

	  // If we're not at the end, set the number for this message
	  if (!bEnd && (i < nMessages)) {
		if (sscanf((LPCTSTR) inputBuffer, "%lu %lu", &(lpdwMessageIDs[i]), &(lpdwMessageLengths[i])) != 2)
		{
		  // Could not chew the line; emit an error
		  nMessages = 0;
		  dwMailboxLength = 0;
 	      pSocket.Close();
		  return EmitErrorMessage("Could not decipher server response to LIST command", 21);
		}
		i++;
	  }
	}

	// Initialize the msgData members to zero
	for (i = 0; i < nMessages; i++)
	  msgData[i] = NULL;
  } else {
	lpdwMessageLengths = NULL;
	lpdwMessageIDs = NULL;
	msgData = NULL;
  }

  // We're in ;-)!
  EmitErrorMessage("Connected", 0);
  return S_OK;
}

STDMETHODIMP CPOP3Connection::Close()
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())

  // If we're not connected, then return an error
  if (!connected)
	return EmitErrorMessage("Connection already closed", 19);

  // Send the QUIT command to disconnect
  CString pData;
  pData = "QUIT\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send QUIT to the server", 20);
  }

  // Eat up the response, then close (we're being polite to the server)
  BOOL bOk;
  ReadResponse(&bOk);

  // Close the connection
  pSocket.Close();
  EmitErrorMessage("Connection closed", 0);
  connected = false;
  return S_OK;
}

/*  STAT		MessageCount and MailboxLength properties
    LIST		MessageLength(msg)
    RETR msg	Message(msg)
    DELE msg	Delete msg
	NOOP		Idle
	RSET		Undelete */

STDMETHODIMP CPOP3Connection::get_MessageCount(long * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  *pVal = (long) nMessages;
  return S_OK;
}

STDMETHODIMP CPOP3Connection::get_MailboxLength(long * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  *pVal = (long) dwMailboxLength;
  return S_OK;
}

STDMETHODIMP CPOP3Connection::get_Message(long MessageNumber, BSTR * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;

  // Check to see if the message exists
  if (((DWORD) MessageNumber <= nMessages) && ((DWORD) MessageNumber > 0) && (lpdwMessageIDs != NULL))
  {
	// If the message was already retrieved, then return it
	if (msgData[MessageNumber - 1]) {
	  *pVal = T2BSTR((LPCTSTR) (*msgData[MessageNumber - 1]));
	  return S_OK;
	}

	// Fetch the message from the server
	LPTSTR lpszData = (LPTSTR) malloc(30);
	BOOL bOk;
	sprintf(lpszData, "RETR %lu\r\n", lpdwMessageIDs[MessageNumber - 1]);
    if (pSocket.Send(lpszData, strlen(lpszData), 0) == SOCKET_ERROR) {
	  pSocket.Close();
	  return EmitErrorMessage("Could not send RETR to the server", 50);
	}
	free(lpszData);
    if (!ReadResponse(&bOk)) {
	  pSocket.Close();
	  return EmitErrorMessage("Server returned unexpected data on RETR command", 51);
	}
	if (!bOk) {
	  return EmitErrorMessage((LPCTSTR) inputBuffer, 52);
	}

	// Allocate buffer
	msgData[MessageNumber - 1] = new CString();

	// Read the message one line at a time
	BOOL bEnd = false;
	while (!bEnd) {
	  if (!ReadServerLine(&bEnd)) {
	    pSocket.Close();
	    return EmitErrorMessage("Server returned unexpected data on RETR command data", 53);
	  }

	  // Append this line to our message if it's not the end of a message
	  if (!bEnd) {
	    (*(msgData[MessageNumber - 1])) += (LPCTSTR) inputBuffer;
	    (*(msgData[MessageNumber - 1])) += "\r\n";
	  }
	}

	// Give the message to the DCOM client
    *pVal = T2BSTR((LPCTSTR) (*msgData[MessageNumber - 1]));
    EmitErrorMessage("Message downloaded", 0);
  } else
  {
	// Message doesn't exist!
    EmitErrorMessage("Index out of range", 58);
  }
  return S_OK;
}

STDMETHODIMP CPOP3Connection::get_MessageLength(long MessageNumber, long * pVal)
{
  // Get the message length
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  if (((DWORD) MessageNumber <= nMessages) && ((DWORD) MessageNumber > 0) && (lpdwMessageLengths != NULL)) {
	*pVal = lpdwMessageLengths[MessageNumber - 1];
  } else
	EmitErrorMessage("Index out of range", 21);
  EmitErrorMessage("Message length copied", 0);
  return S_OK;
}

STDMETHODIMP CPOP3Connection::Delete(long MessageNumber)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())

  // Check to see if the message exists
  if (((DWORD) MessageNumber <= nMessages) && ((DWORD) MessageNumber > 0) && (lpdwMessageIDs != NULL))
  {
	// Delete message from local cache
	if (msgData[MessageNumber - 1] != NULL) {
	  delete msgData[MessageNumber - 1];
	  msgData[MessageNumber - 1] = NULL;
	}

    // Delete a message from the server
	LPTSTR lpszData = (LPTSTR) malloc(30);
	BOOL bOk;
	sprintf(lpszData, "DELE %lu\r\n", lpdwMessageIDs[MessageNumber - 1]);
    if (pSocket.Send(lpszData, strlen(lpszData), 0) == SOCKET_ERROR) {
	  pSocket.Close();
	  return EmitErrorMessage("Could not send DELE to the server", 60);
	}
	free(lpszData);
    if (!ReadResponse(&bOk)) {
	  pSocket.Close();
	  return EmitErrorMessage("Server returned unexpected data on RETR command", 61);
	}
	if (!bOk) {
	  return EmitErrorMessage((LPCTSTR) inputBuffer, 62);
	}
    EmitErrorMessage("Message deleted", 0);
  } else
    EmitErrorMessage("Index out of range", 63);
  return S_OK;
}

STDMETHODIMP CPOP3Connection::Idle()
{
  // Send a NOOP command to stay active
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  CString pData;
  pData = "NOOP\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send NOOP to the server", 70);
  }

  // Eat up the response, then close (we're being polite to the server)
  BOOL bOk;
  ReadResponse(&bOk);
  EmitErrorMessage("Connection is idle but active", 0);
  return S_OK;
}

STDMETHODIMP CPOP3Connection::Undelete()
{
  // Undelete any deleted messages
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  CString pData;
  pData = "RSET\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send RSET to the server", 80);
  }

  // Eat up the response, then close (we're being polite to the server)
  BOOL bOk;
  ReadResponse(&bOk);
  EmitErrorMessage("Deleted messages have been undeleted", 0);
  return S_OK;
}

STDMETHODIMP CPOP3Connection::Uncache(long MessageNumber)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())

  // Check to see if the message exists; remove it from the cache if it's there
  if (((DWORD) MessageNumber <= nMessages) && ((DWORD) MessageNumber > 0)) {
	if (msgData[MessageNumber - 1] != NULL)
	{
	  // Uncache!
	  delete msgData[MessageNumber - 1];
	  msgData[MessageNumber - 1] = NULL;
      EmitErrorMessage("Message removed from cache", 0);
	} else
      EmitErrorMessage("Message not in cache", 0);
  } else
    EmitErrorMessage("Index out of range", 90);
  return S_OK;
}
