// SMTPMailer.cpp : Implementation of CSMTPMailer
#include "stdafx.h"
#include "ActiveMail.h"
#include "SMTPMailer.h"

/////////////////////////////////////////////////////////////////////////////
// CSMTPMailer



BOOL CSMTPMailer::EmitErrorMessage(LPCTSTR message, int code)
{
  // Store the error message and return S_OK
  free(ErrorMessage);
  ErrorMessage = (char *) malloc(lstrlen(message) + 1);
  lstrcpy(ErrorMessage, message);
  ErrorCode = code;
  return S_OK;
}

int CSMTPMailer::GetResponseCode(CSocket *pSocket) {
  char nextChar = ' ';
  char inputBuffer[256];
  int nextIndex = 0;
  
  while ((nextChar != '\n') && (nextIndex < 255)) {
    if (pSocket->Receive(&nextChar, 1, 0) == SOCKET_ERROR)
	  return -1;
	inputBuffer[nextIndex++] = nextChar;
  }
  inputBuffer[nextIndex] = '\0';
  
  // This is to support multi-line responses, common with
  // servers that speak ESMTP
  if(inputBuffer[3] == '-' )
	return GetResponseCode(pSocket);
  else
    return atoi(inputBuffer);
}

BOOL CSMTPMailer::ReadLine(CFile *pFile, CString *inputBuffer) {
  // Read the line
  char nextChar;
  int nextIndex = 0;
  *inputBuffer = "";
  if (pFile->Read(&nextChar, 1) < 1)
    return FALSE;
  while (nextChar != '\n') {
	*inputBuffer += nextChar;
	nextIndex++;
    if (pFile->Read(&nextChar, 1) < 1)
	  return FALSE;
  }
  if (inputBuffer->GetAt(nextIndex - 1) == '\r')
	inputBuffer->SetAt(nextIndex - 1, 0);
  return TRUE;
}

// Reads a list from inputBuffer and consumes the string, returning a list in *list
BOOL CSMTPMailer::ReadList(CString *inputBuffer, CObArray *list) {
  int index;
  CString currentAddress;

  // Make sure this string isn't empty
  inputBuffer->TrimLeft();
  inputBuffer->TrimRight();
  if (*inputBuffer == "")
	return TRUE;
  
  // Loop for the list; each item is processed one at a time
  do {
	// Get the current address block
	index = inputBuffer->Find(',');
	if (index >= 0) {
	  inputBuffer->TrimRight();
	  if (index == inputBuffer->GetLength() - 1) {
		// We're at the end of the line and have more data! -- return TRUE
		return TRUE;
	  }	else {		
		currentAddress = inputBuffer->Left(index);
		*inputBuffer = inputBuffer->Mid(index + 1);
	  }
	} else {
	  currentAddress = *inputBuffer;
	  *inputBuffer = "";
	}
	
	// See if there's any <>s
	if ((index = currentAddress.Find('<')) >= 0) {
	  // Get the address in the <>s
	  currentAddress = currentAddress.Mid(index + 1, currentAddress.Find('>') - index - 1);
	}
	
	// Trim the edges of the address for whitespace
	currentAddress.TrimLeft();
	currentAddress.TrimRight();
	
	// Write this address to the list
    list->Add((CObject *) new CString(currentAddress));
  } while (*inputBuffer != "");
  
  // Return FALSE to indicate no more items required
  return FALSE;
}

BOOL CSMTPMailer::ReadListFromFile(CFile *pFile, CString *inputBuffer, CObArray *list) {
  // Read the strings until we have come to the end of this parameter :-)!
  while (ReadList(inputBuffer, list)) {
	// Get the next line.  We've come to the end of this one and there's a comma
	if (!ReadLine(pFile, inputBuffer))
	  return FALSE;
  }
  return TRUE;
}

BOOL CSMTPMailer::SendFromAndTo(CSocket *pSocket, CString *from, CObArray *to) {
  // Format the from variable, just in case
  int index;
  from->TrimLeft();
  from->TrimRight();
  if ((index = from->Find('<')) >= 0) {
	// Get the address in the <>s
	*from = from->Mid(index + 1, from->Find('>') - index - 1);
    from->TrimLeft();
	from->TrimRight();
  }
  
  // Send the MAIL FROM command
  CString pData;
  int code;
  pData = "MAIL FROM:<";
  pData += (*from);
  pData += ">\r\n";
  if (pSocket->Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR)
	return EmitErrorMessage("Could not send MAIL FROM to the server", 2);
  if ((code = GetResponseCode(pSocket)) != 250)
	return EmitErrorMessage("Server returned MAIL FROM error", code);
  
  // Send the RCPT TO command for each recipient
  int len = to->GetSize();
  int i;
  for (i = 0; i < len; i++) {
    pData = "RCPT TO:<";
    pData += (LPCTSTR) *((CString *) (*to)[i]);
    pData += ">\r\n";
    if (pSocket->Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	  EmitErrorMessage("Could not send RCPT TO to the server", 3);
	  return FALSE;
	}
    if ((code = GetResponseCode(pSocket)) != 250) {
	  EmitErrorMessage("Server returned RCPT TO error", code);
	  return FALSE;
	}
	delete (CString *) (*to)[i];
  }

  // Processed well :-)
  return TRUE;
}

BOOL CSMTPMailer::ParseRawHeader(CFile *pFile, CSocket *pSocket, DWORD *excludeStart, DWORD *excludeEnd) {
  // Parses a raw header by telling the SMTP server to send the message
  // to those in the Cc and Bcc fields
  CString from;
  CObArray to;
  CObArray ignored;
  CString inputBuffer;
  BOOL firstLine = TRUE;
  DWORD filePos = 0;
  
  // Set the initial exclude positions
  *excludeStart = -1;
  *excludeEnd = -1;
  
  // Read the line
  ReadLine(pFile, &inputBuffer);
  
  // Line parser -- loop until we reach a blank line, or the end of the header
  do {
	// To element
	if (inputBuffer.Left(3).CompareNoCase("to:") == 0) {
	  inputBuffer = inputBuffer.Right(inputBuffer.GetLength() - 3);

	  // If we are to not send to the To people, then put it in the ignored list
	  if (IgnoreTo)
		ReadListFromFile(pFile, &inputBuffer, &ignored);
	  else
		ReadListFromFile(pFile, &inputBuffer, &to);

	  
	  // Read the line
	  firstLine = FALSE;
	  filePos = pFile->GetPosition();
      ReadLine(pFile, &inputBuffer);
      continue;
	}
	
	// From element
	if (inputBuffer.Left(5).CompareNoCase("from:") == 0) {
      // Set the mode and parse the From line
      int index;
	  inputBuffer = inputBuffer.Right(inputBuffer.GetLength() - 3);
	  inputBuffer.TrimLeft();
	  inputBuffer.TrimRight();
	  
	  if ((index = inputBuffer.Find('<')) >= 0) {
		// Get the address in the <>s
		from = inputBuffer.Mid(index + 1, inputBuffer.Find('>') - index - 1);
        from.TrimLeft();
		from.TrimRight();
	  } else
		from = inputBuffer;
	  
	  // Read the line
	  firstLine = FALSE;
	  filePos = pFile->GetPosition();
      ReadLine(pFile, &inputBuffer);
	  continue;
	}
	
	// Carbon copy element
	if (inputBuffer.Left(3).CompareNoCase("cc:") == 0) {
	  inputBuffer = inputBuffer.Right(inputBuffer.GetLength() - 3);
	  ReadListFromFile(pFile, &inputBuffer, &to);
	  
	  // Read the line
	  firstLine = FALSE;
	  filePos = pFile->GetPosition();
      ReadLine(pFile, &inputBuffer);
	  continue;
	}
	
	// Blind carbon copy element
	if (inputBuffer.Left(4).CompareNoCase("bcc:") == 0) {
	  inputBuffer = inputBuffer.Right(inputBuffer.GetLength() - 4);
	  ReadListFromFile(pFile, &inputBuffer, &to);
	  
	  // Exclude from the file
	  *excludeStart = filePos;
	  *excludeEnd = pFile->GetPosition();
	  
	  // Read the line
	  firstLine = FALSE;
	  filePos = pFile->GetPosition();
      ReadLine(pFile, &inputBuffer);
	  continue;
	}
	
	// Not yet parsed!  Perform alternate methods to determine if this is still a header
	
	// Blank line: end of header
	if (inputBuffer != "") {
	  // Not a header element!  Return FALSE if this isn't the first
	  // line (raw messages _must_ have a header!)
	  if (firstLine && (inputBuffer.Find(':') == -1)) {
		EmitErrorMessage("Message does not have a header", 1);
		return FALSE;
	  }
	}
	
	// Read the line
    ReadLine(pFile, &inputBuffer);
  } while (inputBuffer != "");
  
  
  // Now that we have the e-mail addresses in memory, let's set up the SMTP server
  // to send to them =)!
  if (!SendFromAndTo(pSocket, &from, &to))
	return FALSE;
  
  // Everything completed successfully =)
  pFile->SeekToBegin();
  return TRUE;
}

STDMETHODIMP CSMTPMailer::SendRawMessage(BSTR Message)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  EmitErrorMessage("SendRawMessage not implemented in this beta", 25);
  return S_OK;
}

STDMETHODIMP CSMTPMailer::SendRawMessageFromFile(BSTR FileName)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  
  // Open the file first
  CFile pFile;
  LPTSTR lpszFileName = OLE2T(FileName);
  int code;
  if (!pFile.Open(lpszFileName, CFile::modeRead))
	return EmitErrorMessage("Cannot open file", 4);
  
  // Connect to the SMTP server
  CSocket pSocket;
  pSocket.Create(0, SOCK_STREAM, 0);
  
  if (!pSocket.Connect(lpszServerName, 25)) {
	if (pSocket.GetLastError() == WSAENOTSOCK) {
	  pSocket.Close();
	  pFile.Close();
	  return EmitErrorMessage("Connection would block", 5);
	} else {
	  pFile.Close();
 	  return EmitErrorMessage("Could not connect to server", 6);
	}
  }
  if ((code = GetResponseCode(&pSocket)) != 220) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned incorrect response on connection", code);
  }

  // Send the HELO command
  CString pData;
  char hostName[256];
  pData = "HELO ";
  if (!gethostname(hostName, 256))
	strcpy(hostName, "localhost");
  pData += hostName;
  pData += "\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Could not send HELO to the server", 7);
  }
  if ((code = GetResponseCode(&pSocket)) != 250) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned HELO error", code);
  }
  
  // Parse the raw header in order to extract the From, To, Cc, and Bcc fields
  DWORD excludeStart, excludeEnd;
  if (!ParseRawHeader(&pFile, &pSocket, &excludeStart, &excludeEnd)) {
	pSocket.Close();
	pFile.Close();
	return S_OK;
  }
  
  // Send the DATA command
  pData = "DATA\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Could not send DATA to the server", 8);
  }
  if ((code = GetResponseCode(&pSocket)) != 354) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned DATA error", code);
  }
  
  // Write the contents of the file to the server
  UINT bytesRead;
  DWORD filePos = 0;
  char fileBuffer[1024];
  try {
    do {
	  if ((filePos >= excludeStart) && (filePos < excludeEnd)) {
		CString *temp = new CString();
		temp->GetBufferSetLength(excludeEnd - excludeStart + 2);
		bytesRead = pFile.Read(temp->LockBuffer(), excludeEnd - excludeStart);
		temp->UnlockBuffer();
		delete temp;
	  } else {
		if (((filePos + 1024 >= excludeStart) && (filePos + 1024 < excludeEnd)) ||
		  ((filePos <= excludeStart) && (filePos + 1024 > excludeStart)))
		  bytesRead = pFile.Read(fileBuffer, excludeStart - filePos);
		else
          bytesRead = pFile.Read(fileBuffer, 1024);
		if (bytesRead) {
		  if (pSocket.Send(fileBuffer, bytesRead, 0) == SOCKET_ERROR) {
	        pSocket.Close();
	        pFile.Close();
			return EmitErrorMessage("Could not write message to the server", 10);
		  }
		}
	  }
	  filePos += bytesRead;
	} while (bytesRead > 0);
  } catch (CFileException) {};
  
  // Write ending
  if (pSocket.Send("\r\n.\r\n", 5) == SOCKET_ERROR) {
	pSocket.Close();
	pFile.Close();
    return EmitErrorMessage("Could not write message ending to the server", 11);
  }
  
  if ((code = GetResponseCode(&pSocket)) != 250) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned message error", code);
  }
  
  // Send the QUIT command
  pData = "QUIT\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Could not send QUIT to the server", 12);
  }
  if ((code = GetResponseCode(&pSocket)) != 221) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned QUIT error", code);
  }
  
  // Close the socket and file
  pSocket.Close();
  pFile.Close();
  return EmitErrorMessage("Message sent", 0);
}

STDMETHODIMP CSMTPMailer::SendMessageFromFile(BSTR From, BSTR To, BSTR Cc, BSTR Bcc, BSTR Subject, BSTR FileName)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;

  // Format body of message
  LPTSTR lpszFrom = OLE2T(From);
  LPTSTR lpszTo = OLE2T(To);
  LPTSTR lpszCc = OLE2T(Cc);
  LPTSTR lpszBcc = OLE2T(Bcc);
  LPTSTR lpszSubject = OLE2T(Subject);
  CString body;
  if (lstrlen(lpszFrom) > 0)
    body = (CString("From: ") + lpszFrom) + "\r\n";
  else
	return EmitErrorMessage("From parameter required", 50);
  if (lstrlen(lpszTo) > 0)
    body += (CString("To: ") + lpszTo) + "\r\n";
  else
	return EmitErrorMessage("To parameter required", 51);
  if (lstrlen(lpszCc) > 0)
    body += (CString("Cc: ") + lpszCc) + "\r\n";
  if (lstrlen(lpszSubject) > 0)
    body += (CString("Subject: ") + lpszSubject) + "\r\n";
  body += "\r\n";

  // Make sure that if we're ignoring To, that we have a Cc or Bcc
  if (IgnoreTo) {
	if ((lstrlen(lpszCc) == 0) && (lstrlen(lpszBcc) == 0))
	  return EmitErrorMessage("To is ignored, and no other destinations given", 55);
  }

  // Process the To, Cc, and Bcc lines into the to array
  CObArray to;
  if (!IgnoreTo)
	ReadList(&CString(lpszTo), &to);
  ReadList(&CString(lpszCc), &to);
  ReadList(&CString(lpszBcc), &to);

  // Open the file
  CFile pFile;
  LPTSTR lpszFileName = OLE2T(FileName);
  if (!pFile.Open(lpszFileName, CFile::modeRead))
	return EmitErrorMessage("Cannot open file", 4);
  
  // Connect to the SMTP server
  CSocket pSocket;
  int code;
  pSocket.Create(0, SOCK_STREAM, 0);
  
  if (!pSocket.Connect(lpszServerName, 25)) {
	if (pSocket.GetLastError() == WSAENOTSOCK) {
	  pFile.Close();
	  pSocket.Close();
	  return EmitErrorMessage("Connection would block", 5);
	} else {
	  pFile.Close();
 	  return EmitErrorMessage("Could not connect to server", 6);
	}
  }
  if ((code = GetResponseCode(&pSocket)) != 220) {
	pFile.Close();
	pSocket.Close();
	return EmitErrorMessage("Server returned incorrect response on connection", code);
  }
  
  // Send the HELO command
  CString pData;
  char hostName[256];
  pData = "HELO ";
  if (!gethostname(hostName, 256))
	strcpy(hostName, "localhost");
  pData += hostName;
  pData += "\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Could not send HELO to the server", 7);
  }
  if ((code = GetResponseCode(&pSocket)) != 250) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned HELO error", code);
  }
  
  // Send the To and From
  if (!SendFromAndTo(&pSocket, &CString(lpszFrom), &to)) {
	pSocket.Close();
	pFile.Close();
	return S_OK;
  }
  
  // Send the DATA command
  pData = "DATA\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Could not send DATA to the server", 8);
  }
  if ((code = GetResponseCode(&pSocket)) != 354) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned DATA error", code);
  }

  // Remove any "\r\n.\r\n" sequences
  int pos;
  while ((pos = body.Find("\r\n.\r\n")) >= 0) {
	body.SetAt(pos + 3, ' ');
  }

  // Write the contents of the buffer to the server
  LPSTR bodyBuffer = body.LockBuffer();
  if (pSocket.Send(bodyBuffer, lstrlen(bodyBuffer), 0) == SOCKET_ERROR) {
	body.UnlockBuffer();
    pSocket.Close();
	pFile.Close();
    return EmitErrorMessage("Could not write message to the server", 10);
  }
  body.UnlockBuffer();

  // Write the contents of the file to the server
  UINT bytesRead;
  char fileBuffer[1024];
  try {
    do {
      bytesRead = pFile.Read(fileBuffer, 1024);
	  if (bytesRead) {
	    if (pSocket.Send(fileBuffer, bytesRead, 0) == SOCKET_ERROR) {
	      pSocket.Close();
	      pFile.Close();
	      return EmitErrorMessage("Could not write message to the server", 10);
		}
	  }
	} while (bytesRead > 0);
  } catch (CFileException) {};
  
  // Write ending
  if (pSocket.Send("\r\n.\r\n", 5) == SOCKET_ERROR) {
	pSocket.Close();
	pFile.Close();
    return EmitErrorMessage("Could not write message ending to the server", 11);
  }
  
  if ((code = GetResponseCode(&pSocket)) != 250) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned message error", code);
  }
  
  // Send the QUIT command
  pData = "QUIT\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Could not send QUIT to the server", 12);
  }
  if ((code = GetResponseCode(&pSocket)) != 221) {
	pSocket.Close();
	pFile.Close();
	return EmitErrorMessage("Server returned QUIT error", code);
  }

  // Close the socket and file
  pSocket.Close();
  pFile.Close();
  return EmitErrorMessage("Message sent", 0);
}

STDMETHODIMP CSMTPMailer::SendMessage(BSTR From, BSTR To, BSTR Cc, BSTR Bcc, BSTR Subject, BSTR Message)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;

  // Format body of message
  LPTSTR lpszFrom = OLE2T(From);
  LPTSTR lpszTo = OLE2T(To);
  LPTSTR lpszCc = OLE2T(Cc);
  LPTSTR lpszBcc = OLE2T(Bcc);
  LPTSTR lpszSubject = OLE2T(Subject);
  LPTSTR lpszMessage = OLE2T(Message);
  CString body;
  if (lstrlen(lpszFrom) > 0)
    body = (CString("From: ") + lpszFrom) + "\r\n";
  else
	return EmitErrorMessage("From parameter required", 50);
  if (lstrlen(lpszTo) > 0)
    body += (CString("To: ") + lpszTo) + "\r\n";
  else
	return EmitErrorMessage("To parameter required", 51);
  if (lstrlen(lpszCc) > 0)
    body += (CString("Cc: ") + lpszCc) + "\r\n";
  if (lstrlen(lpszSubject) > 0)
    body += (CString("Subject: ") + lpszSubject) + "\r\n";
  body += "\r\n";
  body += lpszMessage;

  // Make sure that if we're ignoring To, that we have a Cc or Bcc
  if (IgnoreTo) {
	if ((lstrlen(lpszCc) == 0) && (lstrlen(lpszBcc) == 0))
	  return EmitErrorMessage("To is ignored, and no other destinations given", 55);
  }

  // Process the To, Cc, and Bcc lines into the to array
  CObArray to;
  if (!IgnoreTo)
	ReadList(&CString(lpszTo), &to);
  ReadList(&CString(lpszCc), &to);
  ReadList(&CString(lpszBcc), &to);

  // Connect to the SMTP server
  CSocket pSocket;
  int code;
  pSocket.Create(0, SOCK_STREAM, 0);
  
  if (!pSocket.Connect(lpszServerName, 25)) {
	if (pSocket.GetLastError() == WSAENOTSOCK) {
	  pSocket.Close();
	  return EmitErrorMessage("Connection would block", 5);
	} else
 	  return EmitErrorMessage("Could not connect to server", 6);
  }
  if ((code = GetResponseCode(&pSocket)) != 220) {
	pSocket.Close();
	return EmitErrorMessage("Server returned incorrect response on connection", code);
  }
  
  // Send the HELO command
  CString pData;
  char hostName[256];
  pData = "HELO ";
  if (!gethostname(hostName, 256))
	strcpy(hostName, "localhost");
  pData += hostName;
  pData += "\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send HELO to the server", 7);
  }
  if ((code = GetResponseCode(&pSocket)) != 250) {
	pSocket.Close();
	return EmitErrorMessage("Server returned HELO error", code);
  }
  
  // Send the To and From
  if (!SendFromAndTo(&pSocket, &CString(lpszFrom), &to)) {
	pSocket.Close();
	return S_OK;
  }
  
  // Send the DATA command
  pData = "DATA\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send DATA to the server", 8);
  }
  if ((code = GetResponseCode(&pSocket)) != 354) {
	pSocket.Close();
	return EmitErrorMessage("Server returned DATA error", code);
  }

  // Remove any "\r\n.\r\n" sequences
  int pos;
  while ((pos = body.Find("\r\n.\r\n")) >= 0) {
	body.SetAt(pos + 3, ' ');
  }

  // Write the contents of the buffer to the server
  LPSTR bodyBuffer = body.LockBuffer();
  if (pSocket.Send(bodyBuffer, lstrlen(bodyBuffer), 0) == SOCKET_ERROR) {
	body.UnlockBuffer();
    pSocket.Close();
    return EmitErrorMessage("Could not write message to the server", 10);
  }
  body.UnlockBuffer();
  
  // Write ending
  if (pSocket.Send("\r\n.\r\n", 5) == SOCKET_ERROR) {
	pSocket.Close();
    return EmitErrorMessage("Could not write message ending to the server", 11);
  }
  
  if ((code = GetResponseCode(&pSocket)) != 250) {
	pSocket.Close();
	return EmitErrorMessage("Server returned message error", code);
  }
  
  // Send the QUIT command
  pData = "QUIT\r\n";
  if (pSocket.Send((LPCTSTR) pData, pData.GetLength(), 0) == SOCKET_ERROR) {
	pSocket.Close();
	return EmitErrorMessage("Could not send QUIT to the server", 12);
  }
  if ((code = GetResponseCode(&pSocket)) != 221) {
	pSocket.Close();
	return EmitErrorMessage("Server returned QUIT error", code);
  }

  // Close the socket and file
  pSocket.Close();
  return EmitErrorMessage("Message sent", 0);
}

STDMETHODIMP CSMTPMailer::get_IgnoreTo(BOOL * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  *pVal = IgnoreTo;
  return S_OK;
}

STDMETHODIMP CSMTPMailer::put_IgnoreTo(BOOL newVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  IgnoreTo = newVal;
  return S_OK;
}

STDMETHODIMP CSMTPMailer::get_ErrorCode(long * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  *pVal = ErrorCode;
  return S_OK;
}

STDMETHODIMP CSMTPMailer::get_ErrorMessage(BSTR * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  *pVal = T2BSTR(ErrorMessage);
  return S_OK;
}

STDMETHODIMP CSMTPMailer::get_ServerName(BSTR * pVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  *pVal = T2BSTR(ErrorMessage);
  return S_OK;
}

STDMETHODIMP CSMTPMailer::put_ServerName(BSTR newVal)
{
  AFX_MANAGE_STATE(AfxGetStaticModuleState())
  USES_CONVERSION;
  LPSTR lpszNewVal = OLE2T(newVal);
  free(lpszServerName);
  lpszServerName = (char *) malloc(lstrlen(lpszNewVal) + 1);
  lstrcpy(lpszServerName, lpszNewVal);
  return S_OK;
}
